/**
 * JASMINe
 * Copyright (C) 2005-2007 Bull S.A.S.
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JmxAP.java 41 2007-12-13 14:17:00Z tokmensa $
 * --------------------------------------------------------------------------
 */
package org.ow2.jasmine.monitoring.mbeancmd;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.TreeMap;

import javax.management.MBeanServerConnection;

/**
 * A JMX Access Point.
 */
public class JmxAP {
    /**
     * Creates a JmxAP that connects to a URL.
     *
     * @param url  URL to connect to.
     */
    public JmxAP(final String url) {
        j2eeJmxUrl = url;
    }

    /**
     * Creates a JmxAP that connects to the default URL.
     */
    public JmxAP() {
        j2eeJmxUrl = (String) jmxTargets.get(defaultTarget);
    }

    /**
     * @return  The String version of the JMX URL connected to.
     */
    public String getJmxUrlASString() {
        return j2eeJmxUrl;
    }

    /**
     * @return  The MBeanServerConnection instance used to connect to the J2EE
     *          server via JMX.
     */
    public MBeanServerConnection getMBeanServerConnection() {
        jmxCnx = new JmxCnxWrapper(j2eeJmxUrl);
        return jmxCnx;
    }

    /**
     * @param propertyFile  File name of the properties file.
     *
     * @return  Parsed version of the propertyFile file.
     *
     * @throws IOException  If propertyFile cannot be read.
     */
    private static Properties getProperties(final File propertyFile) throws IOException {
        Properties properties = new Properties();

        FileInputStream in = new FileInputStream(propertyFile);
        properties.load(in);
        in.close();

        return properties;
    }

    /**
     * Gets the JMX URL for a given server name.
     *
     * @param name  Name of the server.
     *
     * @return  JMX URL corresponding to that server name.
     */
    public static String getJmxUrl(final String name) {
        return (String) jmxTargets.get(name);
    }

    /**
     * @return  An array containing the names of all current JMX targets.
     */
    public static String[] getJmxTargets() {

        String[] targets = null;
        if (!jmxTargets.isEmpty()) {
            targets = new String[jmxTargets.size()];
            targets = (String[]) jmxTargets.keySet().toArray(targets);
        }
        return targets;
    }

    /**
     * Gets the names of all JMX targets matching a certain criteria.
     *
     * @param proposed  Criteria.
     *
     * @return  An array containing the names of all current JMX targets.
     */
    public static String[] getJmxTargets(final String[] proposed) {

        boolean all = false;
        String[] targets = null;
        List list = new LinkedList();

        // Look for defined target
        if (proposed != null) {
            for (int i = 0; i < proposed.length; i++) {
                if ("all".equalsIgnoreCase(proposed[i])) {
                    all = true;
                    break;
                }
                String url = getJmxUrl(proposed[i]);
                if (url != null) {
                    list.add(proposed[i]);
                }
            }
        } else {
            if (defaultTarget != null) {
                list.add(defaultTarget);
            }
        }

        if (all) {
            targets = getJmxTargets();
        } else {
            if (list.size() > 0) {
                targets = new String[list.size()];
                for (int i = 0; i < list.size(); i++) {
                     targets[i] = (String) list.get(i);
                }
            }
        }
        return targets;
    }

    /**
     * @return  Default target.
     */
    public static String getDefaultTarget() {
        return defaultTarget;
    }

    /**
     * Loads all JMX connection properties from a file.
     *
     * @param file  File to load JMX properties from.
     */
    private static void load(final File file) {
        // Default target is derived from the j2ee.jmx.url. It may also be
        // defined in the property file using the j2ee.jmx.default.target
        // property. If both are set, j2ee.jmx.url takes precedence.
        String dftTarget = null;
        String url = System.getProperty("j2ee.jmx.url");
        String name = null;
        if (url != null) {
            name = "jonas";
            if (url.matches(".*connector_.*")) {
                int ix = url.lastIndexOf("connector_");
                name = url.substring(ix + "connector_".length());
            }
            jmxTargets.put(name, url);
            defaultTarget = name;
        }

        // Load property file
        Properties props = null;
        try {
            props = getProperties(new File(System.getProperty(
                    "j2ee.jmx.file", DEFAULT_FILE)));
        } catch (IOException e) {
            System.err
                    .println("WARNING : cannot read default JMX configuration file : " + e);
        }

        if ((props != null) && (props.size() != 0)) {
            // Parse the property file
            Enumeration en = props.propertyNames();
            while (en.hasMoreElements()) {
                String key = (String) en.nextElement();
                if (key.startsWith("j2ee.jmx.url.")) {
                    name = key.substring("j2ee.jmx.url.".length());
                    url = props.getProperty(key);
                    jmxTargets.put(name, url);
                }
                if ("j2ee.jmx.default.target".equals(key)) {
                    /*
                     * If j2ee.jmx.url is not set, j2ee.jmx.default.target
                     * defines the default target, provided that it is the name
                     * of a defined URL Check is done at the completion of the
                     * loop
                     */
                    if (defaultTarget == null) {
                        dftTarget = props.getProperty(key);
                    }
                }
            }
        }

        // Check that the default target belongs to the targets
        if (defaultTarget == null) {
            if (dftTarget != null) {
                if (jmxTargets.get(dftTarget) != null) {
                    defaultTarget = dftTarget;
                } else {
                    System.err
                            .println("WARNING : j2ee.jmx.default.target specifies an invalid default target "
                                    + dftTarget + ". No default target.");
                }
            } else {
                // use default URL as default target
                defaultTarget = "jonas";
                jmxTargets.put(defaultTarget, DEFAULT_URL);
                System.err
                        .println("WARNING : j2ee.jmx.default.target not specified => add the default url : "
                                + defaultTarget + " url=" + DEFAULT_URL);
            }
        }
    }

    /**
     * Current JMX connection.
     */
    private MBeanServerConnection jmxCnx = null;

    /**
     * Current JMX URL.
     */
    private String j2eeJmxUrl = DEFAULT_URL;

    /**
     * Default JMX target.
     */
    private static String defaultTarget = null;

    /**
     * Default JMX URL.
     */
    public static final String DEFAULT_URL = "service:jmx:rmi:///jndi/rmi://localhost:1099/jrmpconnector_jonas";

    /**
     * Default name of the properties file.
     */
    public static final String DEFAULT_FILE = "jmxurls.properties";

    /**
     * Map of JMX targets.
     */
    private static Map jmxTargets = new TreeMap();

    static {
        load(new File(System.getProperty("j2ee.jmx.file", DEFAULT_FILE)));
    }
}
